diff --git a/sys/arm/arm/pmap-v6-new.c b/sys/arm/arm/pmap-v6-new.c
index 13a97de..60dd3a2 100644
--- a/sys/arm/arm/pmap-v6-new.c
+++ b/sys/arm/arm/pmap-v6-new.c
@@ -1456,6 +1456,14 @@ static u_long pmap_pte1_demotions;
 SYSCTL_ULONG(_vm_pmap_pte1, OID_AUTO, demotions, CTLFLAG_RD,
     &pmap_pte1_demotions, 0, "1MB page demotions");
 
+static u_long pmap_pte1_kdemotions;
+SYSCTL_ULONG(_vm_pmap_pte1, OID_AUTO, kdemotions, CTLFLAG_RD,
+    &pmap_pte1_kdemotions, 0, "1MB page kernel demotions");
+
+static u_long pmap_pte1_kpromotions;
+SYSCTL_ULONG(_vm_pmap_pte1, OID_AUTO, kpromotions, CTLFLAG_RD,
+    &pmap_pte1_kpromotions, 0, "1MB page kernel promotions");
+
 static u_long pmap_pte1_mappings;
 SYSCTL_ULONG(_vm_pmap_pte1, OID_AUTO, mappings, CTLFLAG_RD,
     &pmap_pte1_mappings, 0, "1MB page mappings");
@@ -1468,6 +1476,14 @@ static u_long pmap_pte1_promotions;
 SYSCTL_ULONG(_vm_pmap_pte1, OID_AUTO, promotions, CTLFLAG_RD,
     &pmap_pte1_promotions, 0, "1MB page promotions");
 
+static u_long pmap_pte1_useraborts;
+SYSCTL_ULONG(_vm_pmap_pte1, OID_AUTO, useraborts, CTLFLAG_RD,
+    &pmap_pte1_useraborts, 0, "1MB page user fixes");
+
+static u_long pmap_pte1_userfixes;
+SYSCTL_ULONG(_vm_pmap_pte1, OID_AUTO, userfixes, CTLFLAG_RD,
+    &pmap_pte1_userfixes, 0, "1MB page user fixes");
+
 static __inline ttb_entry_t
 pmap_ttb_get(pmap_t pmap)
 {
@@ -3133,6 +3149,142 @@ pmap_pv_insert_pte1(pmap_t pmap, vm_offset_t va, vm_paddr_t pa)
 		return (FALSE);
 }
 
+static void
+pmap_update_pte1_kernel(vm_offset_t va, pt1_entry_t npte1)
+{
+	pmap_t pmap;
+	pt1_entry_t *pte1p;
+
+	/*
+	 * Get current pmap. Interrupts should be disabled here
+	 * so PCPU_GET() is done atomically.
+	 */
+	pmap = PCPU_GET(curpmap);
+	if (pmap == NULL)
+		pmap = kernel_pmap;
+
+	/*
+	 *   (1) Change pte1 on current pmap.
+	 *   (2) Flush all obsolete TLB entries on current CPU.
+	 *   (3) Change pte1 on all pmaps.
+	 *   (4) Flush all obsolete TLB entries on all CPUs in SMP case.
+	 */
+
+	pte1p = pmap_pte1(pmap, va);
+	pte1_store(pte1p, npte1);
+
+	if (pte1_is_section(npte1)) {
+		/* Kill all the small mappings. */
+		tlb_flush_range_local(pte1_trunc(va), PTE1_SIZE);
+		pmap_pte1_kpromotions++;
+	} else {
+		/* Kill the one big mapping. */
+		tlb_flush_local(pte1_trunc(va));
+		pmap_pte1_kdemotions++;
+	}
+
+	/*
+	 * This function is called either when all cpus are at smp rendezvous
+	 * in SMP case or spinlock_enter() <-> spinlock_exit() section is used
+	 * in UP case. Thus there is no need to use 'allpmaps_lock' spinlock
+	 * mutex here.
+	 */
+	LIST_FOREACH(pmap, &allpmaps, pm_list) {
+		pte1p = pmap_pte1(pmap, va);
+		pte1_store(pte1p, npte1);
+	}
+#ifdef SMP
+	if (pte1_is_section(npte1)) {
+		/* Kill all the small mappings. */
+		tlb_flush_range(pte1_trunc(va), PTE1_SIZE);
+		pmap_pte1_kpromotions++;
+	} else {
+		/* Kill the one big mapping. */
+		tlb_flush(pte1_trunc(va));
+		pmap_pte1_kdemotions++;
+	}
+#endif
+}
+
+#ifdef SMP
+struct pte1_action {
+	vm_offset_t va;
+	pt1_entry_t npte1;
+	u_int update;		/* CPU that updates the PTE1 */
+};
+
+static void
+pmap_update_pte1_action(void *arg)
+{
+	struct pte1_action *act = arg;
+
+	if (act->update == PCPU_GET(cpuid))
+		pmap_update_pte1_kernel(act->va, act->npte1);
+}
+#endif
+/*
+ *  Change pte1 on current pmap.
+ *  Note that kernel pte1 must be changed on all pmaps.
+ *
+ *  By ARM ARM manual, the behaviour is UNPREDICABLE when two or more TLB
+ *  entries map same VA. It's only problem when either promotion or demotion
+ *  is being done. It these cases, pte1 update and appropriate TLB flush must
+ *  be done atomically in general.
+ */
+static void
+pmap_change_pte1(pmap_t pmap, pt1_entry_t *pte1p, vm_offset_t va,
+    pt1_entry_t npte1)
+{
+
+	PMAP_LOCK_ASSERT(pmap, MA_OWNED);
+
+	if (pmap == kernel_pmap) {
+		/*
+		 * The break-before-make approach cannot be used for kernel
+		 * mappings. It would be unhappy to unmap itself unexpectedly.
+		 * We need to freeze everything as we are going to change pte1
+		 * of kernel mapping.
+		 */
+#ifdef SMP
+		struct pte1_action act;
+
+		sched_pin();
+		act.va = va;
+		act.npte1 = npte1;
+		act.update = PCPU_GET(cpuid);
+		smp_rendezvous_cpus(all_cpus, smp_no_rendevous_barrier,
+		    pmap_update_pte1_action, NULL, &act);
+		sched_unpin();
+#else
+		spinlock_enter();
+		pmap_update_pte1_kernel(va, npte1);
+		spinlock_exit();
+#endif
+	} else {
+		/*
+		 * We can use break-before-make approach for changing
+		 * userland mappings. This can lead to L1 translation
+		 * aborts, so special treatment is done in pmap_fault().
+		 *
+		 * The spinlock_enter() <-> spinlock_exit() section is used
+		 * here to make pte1 change as quick as possible without any
+		 * interruption. This way no L1 translation abort can happen
+		 * in UP case.
+		 */
+		spinlock_enter();
+		pte1_clear(pte1p);
+		if (pte1_is_section(npte1)) {
+			/* Kill all the small mappings. */
+			pmap_tlb_flush_range(pmap, pte1_trunc(va), PTE1_SIZE);
+		} else {
+			/* Kill the one big mapping. */
+			pmap_tlb_flush(pmap, pte1_trunc(va));
+		}
+		pte1_store(pte1p, npte1);
+		spinlock_exit();
+	}
+}
+
 /*
  *  Tries to promote the NPTE2_IN_PT2, contiguous 4KB page mappings that are
  *  within a single page table page (PT2) to a single 1MB page mapping.
@@ -3261,17 +3413,7 @@ setpte2:
 	/*
 	 * Map the section.
 	 */
-	if (pmap == kernel_pmap)
-		pmap_kenter_pte1(va, npte1);
-	else
-		pte1_store(pte1p, npte1);
-	/*
-	 * Flush old small mappings. We call single pmap_tlb_flush() in
-	 * pmap_demote_pte1() and pmap_remove_pte1(), so we must be sure that
-	 * no small mappings survive. We assume that given pmap is current and
-	 * don't play game with PTE2_NG.
-	 */
-	pmap_tlb_flush_range(pmap, pte1_trunc(va), PTE1_SIZE);
+	pmap_change_pte1(pmap, pte1p, va, npte1);
 
 	pmap_pte1_promotions++;
 	CTR3(KTR_PMAP, "%s: success for va %#x in pmap %p",
@@ -3556,17 +3698,7 @@ pmap_demote_pte1(pmap_t pmap, pt1_entry_t *pte1p, vm_offset_t va)
 	 * another processor changing the setting of PTE1_A and/or PTE1_NM
 	 * between the read above and the store below.
 	 */
-	if (pmap == kernel_pmap)
-		pmap_kenter_pte1(va, npte1);
-	else
-		pte1_store(pte1p, npte1);
-
-	/*
-	 * Flush old big mapping. The mapping should occupy one and only
-	 * TLB entry. So, pmap_tlb_flush() called with aligned address
-	 * should be sufficient.
-	 */
-	pmap_tlb_flush(pmap, pte1_trunc(va));
+	pmap_change_pte1(pmap, pte1p, va, npte1);
 
 	/*
 	 * Demote the pv entry. This depends on the earlier demotion
@@ -6111,9 +6243,17 @@ pmap_fault(pmap_t pmap, vm_offset_t far, uint32_t fsr, int idx, int usermode)
 		 * only permission abort from malicious user is not fatal.
 		 */
 		if (!usermode || (idx != FAULT_PERM_L2)) {
+			uint32_t par;
+
 			CTR4(KTR_PMAP, "%s: pmap %#x pm_pt1 %#x far %#x",
 			    __func__, pmap, pmap->pm_pt1, far);
-			panic("%s: pm_pt1 abort", __func__);
+			//panic("%s: pm_pt1 abort", __func__);
+
+			cp15_ats1cpr_set((vm_offset_t)pmap);
+			isb();
+			par = cp15_par_get();
+			printf("%s: pm_pt1 abort - pmap %p (%x)", __func__, pmap, par);
+			breakpoint();
 		}
 		return (EFAULT);
 	}
@@ -6208,6 +6348,28 @@ pte1_setrw:
 	 *      starting to deal with not fast aborts.
 	 */
 
+#ifdef SMP
+	/*
+	 * Check that the abort is not caused by break-before-make approch
+	 * done in pmap_change_pte1() for userland mappings in SMP case.
+	 */
+	if (idx == FAULT_TRAN_L1 && usermode) {
+		uint32_t par;
+
+		pmap_pte1_useraborts++;
+
+		PMAP_LOCK(pmap);
+		cp15_ats1cur_set(far);
+		isb();
+		par = cp15_par_get();
+		PMAP_UNLOCK(pmap);
+
+		if ((par & 0x01) == 0) {
+			pmap_pte1_userfixes++;
+			return (0);
+		}
+	}
+#endif
 #ifdef INVARIANTS
 	/*
 	 * Read an entry in PT2TAB associated with both pmap and far.
diff --git a/sys/arm/include/cpu-v6.h b/sys/arm/include/cpu-v6.h
index e4a8936..bb2a707 100644
--- a/sys/arm/include/cpu-v6.h
+++ b/sys/arm/include/cpu-v6.h
@@ -155,8 +155,9 @@ _RF0(cp15_l2ctlr_get, CP15_L2CTLR(%0))
 _RF0(cp15_actlr_get, CP15_ACTLR(%0))
 _WF1(cp15_actlr_set, CP15_ACTLR(%0))
 #if __ARM_ARCH >= 6
-_WF1(cp15_ats1cpr_set, CP15_ATS1CPR(%0));
-_RF0(cp15_par_get, CP15_PAR);
+_WF1(cp15_ats1cpr_set, CP15_ATS1CPR(%0))
+_WF1(cp15_ats1cur_set, CP15_ATS1CUR(%0))
+_RF0(cp15_par_get, CP15_PAR(%0))
 _RF0(cp15_sctlr_get, CP15_SCTLR(%0))
 #endif
 
diff --git a/sys/arm/include/sysreg.h b/sys/arm/include/sysreg.h
index 31097e9..dc93869 100644
--- a/sys/arm/include/sysreg.h
+++ b/sys/arm/include/sysreg.h
@@ -130,7 +130,7 @@
 #define	CP15_BPIALLIS		p15, 0, r0, c7, c1,  6 /* Branch predictor invalidate all IS */
 #endif
 
-#define	CP15_PAR		p15, 0, r0, c7, c4,  0 /* Physical Address Register */
+#define	CP15_PAR(rr)		p15, 0, rr, c7, c4,  0 /* Physical Address Register */
 
 #define	CP15_ICIALLU		p15, 0, r0, c7, c5,  0 /* Instruction cache invalidate all PoU */
 #define	CP15_ICIMVAU(rr)	p15, 0, rr, c7, c5,  1 /* Instruction cache invalidate */
